
#include <stdio.h>
#include <map2.h>
#include <1905_tlvs.h>
#include <cmdu.h>
#include "cmdu_validate.h"
#include "debug.h"


bool validate_channel_scan_request(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret = 0;
	struct tlv_policy policy[] = {
		[0] = {
			.type = MAP_TLV_CHANNEL_SCAN_REQ,
			.present = TLV_PRESENT_ONE,
			.minlen = 2,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, policy, 1);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "%s: required TLV missing: " \
				"CHANNEL_SCAN_REQ_TLV\n", __func__);
		return false;
	}

	/* Channel Scan Request TLV */
	if (tv[0][0]) {
		int i, j;
		int offset = 0;
		int num_radio = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[0][0]);

		t = (uint8_t *)tv[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(2) */
		offset += 1;	/* mode */
		num_radio = t[offset];
		offset += 1;	/* num radio */

		for (i = 0; i < num_radio; i++) {
			int num_opclass = 0;

			if ((offset + 6 + 1) > tlv_len)
				return false;

			offset += 6;	/* radio id */
			num_opclass = t[offset];
			offset += 1;	/* num opclass */

			for (j = 0; j < num_opclass; j++) {
				int num_ch = 0;

				if ((offset + 1 + 1) > tlv_len)
					return false;

				offset += 1;	/* class id */
				num_ch = t[offset];
				offset += 1;	/* num_channel */

				if ((offset + num_ch) > tlv_len)
					return false;

				offset += num_ch;	/* channels */
			}
		}
	}

	return true;
}

bool validate_channel_scan_report(struct cmdu_buff *cmdu, struct tlv *tv_tsp[][16],
		struct tlv *tv_scan[], int *num)
{
	int i, j;
	int ret = 0;
	struct tlv_policy d_policy_tsp[] = {
		[0] = {
			.type = MAP_TLV_TIMESTAMP,
			.present = TLV_PRESENT_ONE,
			.minlen = 1
		}
	};
	struct tlv_policy d_policy_scan[] = {
		[0] = {
			.type = MAP_TLV_CHANNEL_SCAN_RES,
			.present = TLV_PRESENT_MORE,
			.minlen = 9
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv_tsp, d_policy_tsp, 1);
	ret |= cmdu_parse_tlv_single(cmdu, tv_scan, d_policy_scan, num);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv_tsp[0][0]) {
		fprintf(stderr, "%s: No TIMESTAMP_TLV received!\n", __func__);
		return false;
	}

	if (!tv_scan[0]) {
		fprintf(stderr, "%s: No RESULT_TLV received!\n", __func__);
		return false;
	}

	/* Timestamp TLV */
	if (tv_tsp[0][0]) {
		int offset = 0;
		int tsp_len = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv_tsp[0][0]);

		t = (uint8_t *)tv_tsp[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		tsp_len = t[offset];
		offset += 1;
		if ((offset + tsp_len) > tlv_len)
			return false;
	}

	/* Channel Scan Result TLV */
	for (i = 0; i < *num; i++) {
		int offset = 0;
		uint8_t status = 0x00;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv_scan[i]);
		int ts_len = 0;
		uint16_t num_nb = 0;

		t = (uint8_t *)tv_scan[i]->data;
		if (!t)
			return false;

		/* offset within provided minlen(9) */
		offset += 6;	/* radio id */
		offset += 1;	/* opclass */
		offset += 1;	/* channel */
		status = t[offset];
		offset += 1;	/* status */

		if (status == CH_SCAN_STATUS_SUCCESS) {

			if ((offset + 1) > tlv_len)
				return false;

			ts_len = t[offset];
			offset += 1;	/* timestamp len */
			if ((offset + ts_len + 4) > tlv_len)
				return false;

			offset += ts_len;	/* timestamp */
			offset += 1;		/* utilization */
			offset += 1;		/* noise */
			num_nb = BUF_GET_BE16(t[offset]);
			offset += 2;		/* num neightbors */

			for (j = 0; j < num_nb; j++) {
				int ssid_len = 0;
				int bw_len = 0;
				uint8_t info = 0x00;

				if ((offset + 7) > tlv_len)
					return false;

				offset += 6;	/* bssid */
				ssid_len = t[offset];
				offset += 1;	/* ssid len */
				if ((offset + ssid_len + 2) > tlv_len)
					return false;

				offset += ssid_len;	/* ssid */
				offset += 1;		/* rcpi */
				bw_len = t[offset];
				offset += 1;		/* bw len */
				if ((offset + bw_len + 1) > tlv_len)
					return false;

				offset += bw_len;	/* bandwidth */
				info = t[offset];
				offset += 1;		/* info */

				if (info & CH_SCAN_RESULT_BSSLOAD_PRESENT) {
					if ((offset + 3) > tlv_len)
						return false;

					offset += 1;	/* channel utilization */
					offset += 2;	/* station count */
				}
			}

			if ((offset + 5) > tlv_len)
				return false;

			offset += 4;	/* total scan duration */
			offset += 1;	/* scan type */
		}
	}

	return true;
}

bool validate_topology_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret = 0;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_AP_OPERATIONAL_BSS,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
		},
		[1] = {
			.type = MAP_TLV_ASSOCIATED_CLIENTS,
			.present = TLV_PRESENT_OPTIONAL_ONE,
			.minlen = 1,
		},
		[2] = {
			.type = MAP_TLV_MULTIAP_PROFILE,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 3);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0] || !tv[2][0]) {
		fprintf(stderr, "required TLV missing: " \
				"AP_OPERATIONAL_BSS or MULTIAP_PROFILE\n");
		return false;
	}

	/* AP Operational BSS TLV */
	if (tv[0][0]) {
		int i, j;
		int offset = 0;
		int num_radio = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[0][0]);

		t = (uint8_t *)tv[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen (1) */
		num_radio = t[offset];
		offset += 1;	/* num radio */

		for (i = 0; i < num_radio; i++) {
			int num_bss = 0;

			if ((offset + 6 + 1) > tlv_len)
				return false;

			offset += 6;	/* radio id */
			num_bss = t[offset];
			offset += 1;	/* num bss */

			for (j = 0; j < num_bss; j++) {
				int ssid_len = 0;

				if ((offset + 6 + 1) > tlv_len)
					return false;

				offset += 6;	/* bssid */
				ssid_len = t[offset];
				offset += 1;	/* ssid len */

				if ((offset + ssid_len) > tlv_len)
					return false;

				offset += ssid_len;	/* ssid */
			}
		}
	}

	/* Associated Client TLV */
	if (tv[1][0]) {
		int i;
		int offset = 0;
		int num_bss = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[1][0]);

		t = (uint8_t *)tv[1][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen (1) */
		num_bss = t[offset];
		offset += 1;	/* num bss */

		for (i = 0; i < num_bss; i++) {
			uint16_t num_cl = 0;

			if ((offset + 6 + 2) > tlv_len)
				return false;

			offset += 6;	/* bssid */
			num_cl = BUF_GET_BE16(t[offset]);
			offset += 2;	/* num client */

			if ((offset + (num_cl * 8)) > tlv_len)
				return false;

			/* client macaddr + conntime */
			offset += (num_cl * 8);
		}
	}

	return true;
}

bool validate_topology_query(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret = 0;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_MULTIAP_PROFILE,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 1);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "%s: NO MULTIAP_PROFILE_TLV received!!\n", __func__);
		return false;
	}

	return true;
}

bool validate_ap_metric_query(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret = 0;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_AP_METRIC_QUERY,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
		},
		[1] = {
			.type = MAP_TLV_AP_RADIO_IDENTIFIER,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 6,
			.maxlen = 6,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 2);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "required TLV missing: " \
				"AP_METRIC_Query\n");
		return false;
	}

	/* AP Metric Query */
	if (tv[0][0]) {
		int offset = 0;
		uint8_t *t = NULL;
		int num_bss = 0;
		uint16_t tlv_len = tlv_length(tv[0][0]);

		t = (uint8_t *)tv[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		num_bss = t[offset];
		offset += 1;	/* num bss */

		if ((offset + 6 * num_bss) > tlv_len)
			return false;
	}

	return true;
}

bool validate_ap_metric_resp(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret = 0;
	int index = 0;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_AP_METRICS,
			.present = TLV_PRESENT_MORE,
			.minlen = 13,
		},
		[1] = {
			.type = MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 34,
			.maxlen = 34,
		},
		[2] = {
			.type = MAP_TLV_ASSOCIATED_STA_LINK_METRICS,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 7,
		},
		[3] = {
			.type = MAP_TLV_AP_EXTENDED_METRICS,
			.present = TLV_PRESENT_MORE,
			.minlen = 30,
			.maxlen = 30,
		},
		[4] = {
			.type = MAP_TLV_RADIO_METRICS,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 10,
			.maxlen = 10,
		},
		[5] = {
			.type = MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 7,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 6);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0] || !tv[3][0]) {
		fprintf(stderr, "%s: required TLV missing: " \
				"AP_METRICS_TLV or AP_EXTENDED_METRICS_TLV\n",
				__func__);
		return false;
	}

	/* AP Metrics TLV */
	index = 0;
	while (tv[0][index] && (index < 16)) {
		int offset = 0;
		uint8_t *t = NULL;
		uint8_t esp_ac = 0x00;
		uint16_t tlv_len = tlv_length(tv[0][index]);

		t = (uint8_t *)tv[0][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(13) */
		offset += 6;	/* bssid */
		offset += 1;	/* channel util */
		offset += 2;	/* num station */
		esp_ac = t[offset];
		offset += 1;	/* ESP */
		offset += 3;	/* ESP_BE */

		if (esp_ac & ESP_AC_BK) {
			if ((offset + 3) > tlv_len)
				return false;

			offset += 3;	/* ESP_BK */
		}

		if (esp_ac & ESP_AC_VO) {
			if ((offset + 3) > tlv_len)
				return false;

			offset += 3;	/* ESP_VO */
		}

		if (esp_ac & ESP_AC_VI) {
			if ((offset + 3) > tlv_len)
				return false;

			offset += 3;	/* ESP_VI */
		}

		index++;
	}

	/* Associated STA Link Metrics TLV */
	index = 0;
	while (tv[2][index] && (index < 16)) {
		int offset = 0;
		int num_bss = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[2][index]);

		t = (uint8_t *)tv[2][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(7) */
		offset += 6;	/* sta mac */
		num_bss = t[offset];
		offset += 1;	/* num bss */

		if ((offset + 19 * num_bss) > tlv_len)
			return false;

		index++;
	}

	/* Associcated STA Extended Link Metrics TLV */
	index = 0;
	while (tv[5][index] && (index < 16)) {
		int offset = 0;
		int num_bss = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[5][index]);

		t = (uint8_t *)tv[5][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(7) */
		offset += 6;	/* sta mac */
		num_bss = t[offset];
		offset += 1;	/* num bss */

		if ((offset + 22 * num_bss) > tlv_len)
			return false;

		index++;
	}

	return true;
}

bool validate_ap_cap_report(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int i, j, k, ret = 0;
	int index = 0;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_AP_CAPABILITY,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
			.maxlen = 1,
		},
		[1] = {
			.type = MAP_TLV_AP_RADIO_BASIC_CAPABILITIES,
			.present = TLV_PRESENT_MORE,
			.minlen = 8,
		},
		[2] = {
			.type = MAP_TLV_AP_HT_CAPABILITIES,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 7,
			.maxlen = 7,
		},
		[3] = {
			.type = MAP_TLV_AP_VHT_CAPABILITIES,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 12,
			.maxlen = 12,
		},
		[4] = {
			.type = MAP_TLV_AP_HE_CAPABILITIES,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 9,
		},
		[5] = {
			.type = MAP_TLV_METRIC_COLLECTION_INTERVAL,
			.present = TLV_PRESENT_ONE,
			.minlen = 4,
			.maxlen = 4,
		},
		[6] = {
			.type = MAP_TLV_CHANNEL_SCAN_CAPABILITY,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
		},
		[7] = {
			.type = MAP_TLV_CAC_CAPABILITY,
			.present = TLV_PRESENT_ONE,
			.minlen = 3,
		},
		[8] = {
			.type = MAP_TLV_PROFILE2_AP_CAPABILITY,
			.present = TLV_PRESENT_ONE,
			.minlen = 4,
			.maxlen = 4,
		},
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 9);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0] || !tv[1][0] || !tv[5][0] ||
			!tv[6][0] || !tv[7][0] || !tv[8][0]) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	/* AP Radio Basic Capability TLV */
	index = 0;
	while (tv[1][index]) {
		int offset = 0;
		int num_op = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[1][index]);

		t = (uint8_t *)tv[1][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(8) */
		offset += 6;	/* radio id */
		offset += 1;	/* max bss */
		num_op = t[offset];
		offset += 1;	/* num opclass */

		for (i = 0; i < num_op; i++) {
			int num_nonop_ch = 0;

			if ((offset + 3) > tlv_len)
				return false;

			offset += 1;	/* class id */
			offset += 1;	/* max txpower */
			num_nonop_ch = t[offset];
			offset += 1;	/* num nonop channel */

			if ((offset + num_nonop_ch) > tlv_len)
				return false;

			offset += num_nonop_ch;
		}

		index++;
	}

	/* AP HE Capability TLV */
	index = 0;
	while (tv[4][index]) {
		int offset = 0;
		int he_len = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[4][index]);

		t = (uint8_t *)tv[4][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(9) */
		offset += 6;	/* radio id */
		he_len = t[offset];
		offset += 1;	/* HE len */

		if ((offset + he_len + 2) > tlv_len)
			return false;

		offset += he_len;
		offset += 2;

		index++;
	}

	/* Channel Scan Capability TLV */
	if (tv[6][0]) {
		int offset = 0;
		uint8_t *t = NULL;
		int num_radio = 0;
		int num_op = 0;
		int num_ch = 0;
		uint16_t tlv_len = tlv_length(tv[6][0]);

		t = (uint8_t *)tv[6][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		num_radio = t[offset];
		offset += 1;	/* num radio */

		for (i = 0; i < num_radio; i++) {
			if ((offset + 12) > tlv_len)
				return false;

			offset += 6;	/* radio id */
			offset += 1;	/* cap */
			offset += 4;	/* min scan interval */
			num_op = t[offset];
			offset += 1;	/* num opclass */

			for (j = 0; j < num_op; j++) {
				if ((offset + 2) > tlv_len)
					return false;

				offset += 1;	/* class id */
				num_ch = t[offset];
				offset += 1;	/* num channel */

				if ((offset + num_ch) > tlv_len)
					return false;

				offset += num_ch;
			}
		}
	}

	/* CAC Capability TLV */
	if (tv[7][0]) {
		int offset = 0;
		uint8_t *t = NULL;
		int num_radio = 0;
		int num_cac = 0;
		int num_op = 0;
		int num_ch = 0;
		uint16_t tlv_len = tlv_length(tv[7][0]);

		t = (uint8_t *)tv[7][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(3) */
		offset += 2;	/* country code */
		num_radio = t[offset];
		offset += 1;	/* num radio */

		for (i = 0; i < num_radio; i++) {
			if ((offset + 7) > tlv_len)
				return false;

			offset += 6;	/* radio id */
			num_cac = t[offset];
			offset += 1;	/* num cac */

			for (j = 0; j < num_cac; j++) {
				if ((offset + 5) > tlv_len)
					return false;

				offset += 1;	/* cac method */
				offset += 3;	/* duration */
				num_op = t[offset];
				offset += 1;	/* num opclass */

				for (k = 0; k < num_op; k++) {
					if ((offset + 2) > tlv_len)
						return false;

					offset += 1;	/* class id */
					num_ch = t[offset];
					offset += 1;	/* num channel */

					if ((offset + num_ch) > tlv_len)
						return false;

					offset += num_ch;
				}
			}
		}
	}

	return true;
}

bool validate_bksta_cap_report(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	int index = 0;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY,
			.present = TLV_PRESENT_OPTIONAL_MORE,
			.minlen = 7,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 1);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	/* Backhaul STA Radio Capability TLV */
	index = 0;
	while (tv[0][index]) {
		int offset = 0;
		uint8_t *t = NULL;
		uint8_t mac_include = 0x00;
		uint16_t tlv_len = tlv_length(tv[0][index]);

		t = (uint8_t *)tv[0][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(7) */
		offset += 6;	/* radio id */
		mac_include = t[offset];
		offset += 1;	/* macaddr include */

		if (mac_include & BSTA_MACADDRESS_INCLUDED) {
			if ((offset + 6) > tlv_len)
				return false;

			offset += 6;	/* bksta mac */
		}

		index++;
	}

	return true;
}

bool validate_operating_channel_report(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	int index = 0;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_OPERATING_CHANNEL_REPORT,
			.present = TLV_PRESENT_MORE,
			.minlen = 8,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 1);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	/* Operating Channel Report TLV */
	index = 0;
	while (tv[0][index]) {
		int offset = 0;
		uint8_t *t = NULL;
		int num_opclass = 0;
		uint16_t tlv_len = tlv_length(tv[0][index]);

		t = (uint8_t *)tv[0][index]->data;
		if (!t)
			return false;

		/* offset within provided minlen(8) */
		offset += 6;	/* radio id */
		num_opclass = t[offset];
		offset += 1;	/* num opclass */

		if ((offset + num_opclass * 2 + 1) > tlv_len)
			return false;

		index++;
	}

	return true;
}

bool validate_autoconfig_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_SUPPORTED_SERVICE,
			.present = TLV_PRESENT_OPTIONAL_ONE,
			.minlen = 1,
		},
		[1] = {
			.type = MAP_TLV_MULTIAP_PROFILE,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
			.maxlen = 1,
		},
		[2] = {
			.type = TLV_TYPE_SUPPORTED_FREQ_BAND,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
			.maxlen = 1,
		},
		[3] = {
			.type = TLV_TYPE_SUPPORTED_ROLE,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
			.maxlen = 1,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 4);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[1][0] || !tv[2][0] || !tv[3][0]) {
		fprintf(stderr, "%s: required TLV missing\n", __func__);
		return false;
	}

	/* Supported Service TLV */
	if (tv[0][0]) {
		int offset = 0;
		int num_serv = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[0][0]);

		t = (uint8_t *)tv[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		num_serv = t[offset];
		offset += 1;	/* num supported service */
		if ((offset + num_serv) > tlv_len)
			return false;

	}

	return true;
}

bool validate_topology_notification(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = TLV_TYPE_AL_MAC_ADDRESS_TYPE,
			.present = TLV_PRESENT_ONE,
			.minlen = 6,
			.maxlen = 6,
		},
		[1] = {
			.type = MAP_TLV_CLIENT_ASSOCIATION_EVENT,
			.present = TLV_PRESENT_OPTIONAL_ONE,
			.minlen = 13,
			.maxlen = 13,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 2);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	return true;
}

bool validate_disassoc_stats(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_STA_MAC_ADDRESS,
			.present = TLV_PRESENT_ONE,
			.minlen = 6,
			.maxlen = 6,
		},
		[1] = {
			.type = MAP_TLV_REASON_CODE,
			.present = TLV_PRESENT_ONE,
			.minlen = 2,
			.maxlen = 2,
		},
		[2] = {
			.type = MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS,
			.present = TLV_PRESENT_ONE,
			.minlen = 34,
			.maxlen = 34,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 3);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if ((!tv[0][0] || !tv[1][0] || !tv[2][0])) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	return true;
}

bool validate_client_cap_query(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_CLIENT_INFO,
			.present = TLV_PRESENT_ONE,
			.minlen = 12,
			.maxlen = 12,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 1);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	return true;
}

bool validate_client_cap_report(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_CLIENT_INFO,
			.present = TLV_PRESENT_ONE,
			.minlen = 12,
			.maxlen = 12,
		},
		[1] = {
			.type = MAP_TLV_CLIENT_CAPABILITY_REPORT,
			.present = TLV_PRESENT_ONE,
			.minlen = 1,
		},
		[2] = {
			.type = MAP_TLV_ERROR_CODE,
			.present = TLV_PRESENT_OPTIONAL_ONE,
			.minlen = 7,
			.maxlen = 7,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 3);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0] || !tv[1][0]) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	return true;
}

bool validate_unassoc_sta_link_metric_query(struct cmdu_buff *cmdu,
		struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_QUERY,
			.present = TLV_PRESENT_ONE,
			.minlen = 2,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 1);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	/* Unassociated STA Link Metrics Query */
	{
		int i, offset = 0;
		int num_ch = 0;
		int num_sta = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[0][0]);

		t = (uint8_t *)tv[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen */
		offset += 1;	/* opclass */
		num_ch = t[offset];
		offset += 1;	/* num channel */

		for (i = 0; i < num_ch; i++) {
			if ((offset + 2) > tlv_len)
				return false;

			offset += 1;	/* channel */
			num_sta = t[offset];
			offset += 1;	/* num_sta */

			if ((offset + 6 * num_sta) > tlv_len)
				return false;

			offset += (6 * num_sta);
		}
	}

	return true;
}

bool validate_unassoc_sta_link_metric_resp(struct cmdu_buff *cmdu,
		struct tlv *tv[][16])
{
	int ret;
	struct tlv_policy d_policy[] = {
		[0] = {
			.type = MAP_TLV_UNASSOCIATED_STA_LINK_METRICS_RESPONSE,
			.present = TLV_PRESENT_ONE,
			.minlen = 2,
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv, d_policy, 1);
	if (ret) {
		fprintf(stderr, "%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0]) {
		fprintf(stderr, "%s: required tlv missing\n", __func__);
		return false;
	}

	/* Unassociated STA Link Metrics Response */
	{
		int offset = 0;
		int num_sta = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv[0][0]);

		t = (uint8_t *)tv[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen */
		offset += 1;	/* opclass */
		num_sta = t[offset];
		offset += 1;	/* num sta */

		if ((offset + num_sta * 12) > tlv_len)
			return false;

		offset += (num_sta * 12);
	}

	return true;
}
