/*
 * collector.c
 * WiFi Data Elements Collector main
 *
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>

#include <easy/easy.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <libubus.h>

#include "timer.h"
#include <cmdu.h>
#include <cmdu_ackq.h>

#include "wfa_delm.h"
#include "decollector.h"
#include "debug.h"
#include "config.h"


void decollector_getter_run(void *arg)
{
	struct decollector_private *p = (struct decollector_private *)arg;
	int num = 0;
	uint8_t *macaddrs = NULL;
	struct wifi_data_element *dm;
	int ret;
	int i;

	if (p->getter_running) {
		fprintf(stderr, "Collector already running...\n");
		return;
	}

	fprintf(stdout, "Start getter...\n");

	ret = decollector_get_deagents(p, &num, &macaddrs);
	if (ret || num == 0) {
		err("%s: num agents = %d\n", __func__, num);
		return;
	}

	dm = calloc(1, sizeof(struct wifi_data_element) +
			num * sizeof(struct wifi_network_device));
	if (!dm) {
		free(macaddrs);
		return;
	}

	get_timestamp(NULL, dm->network.tsp);
	dm->network.num_devices = num;
	p->dmi = dm;
	p->pending = num;
	INIT_LIST_HEAD(&dm->network.ssidlist);

	ret = collector_get_network_params(p, &dm->network);
	if (ret)
		err("%s: Error get network ssids!\n", __func__);

	for (i = 0; i < num; i++) {
		memcpy(&dm->network.device[i].macaddr, &macaddrs[i * 6], 6);
		fprintf(stderr, "%s: [%d] device-macaddr = " MACFMT "\n",
			__func__, i, MAC2STR(dm->network.device[i].macaddr));

		ret |= decollector_collect_node(p, &dm->network.device[i]);
	}

	free(macaddrs);

	fprintf(stdout, "Getter running\n");
	p->getter_running = 1;
}
