/*
 * decollector.h
 * WiFi DataElements Collector private header file.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#ifndef DECOLLECTOR_H
#define DECOLLECTOR_H


#define WIFI_DEVICE_MAX_NUM	4
#define WIFI_IFACE_MAX_NUM	16

struct wifi_netdev {
	char radio[16];
	struct {
		char name[16];
		int mode;
	} iface[WIFI_IFACE_MAX_NUM];
};

enum {
	WIFI_IFACE_FH,
	WIFI_IFACE_BK,
};

#define MAP_PROFILE_1	1
#define MAP_PROFILE_2	2

typedef uint32_t object_t;

#define OBJECT_INVALID 0

#define IEEE1905_OBJECT			"ieee1905"
#define MAP_PLUGIN_OBJECT		"ieee1905.map"
#define WFADATAD_AGENT_OBJECT		"wifi.dataelements"
#define WFADATAD_COLLECTOR_OBJECT	"wifi.dataelements.collector"

struct wifi_assoc_frame {
	uint16_t len;
	uint8_t macaddr[6];
	uint8_t *frame;
	struct list_head list;
};

struct decollector_config {
	uint32_t refresh_int;
	bool is_collector;
	bool is_sap_mode;
	uint32_t profile;
	struct ubus_context *uctx;
	char mapcontroller_conffile[256];
};

#define DELM_OBJECT_PLACEHOLDER			\
	struct ubus_context *ctx;		\
	struct ubus_object obj;			\
	struct ubus_event_handler topology_ev;	\
	struct ubus_event_handler sta_ev;	\
	struct ubus_event_handler ubus_obj_ev;	\
	struct ubus_event_handler wifi_radio_ev;


struct decollector_private {
	struct wifi_netdev ifs[WIFI_DEVICE_MAX_NUM];
	struct decollector_config cfg;
	atimer_t t;
	int getter_running;
	uint8_t cntlr_id[6];		/* controller id */
	bool dm_available;
	struct wifi_data_element *dm;
	struct wifi_data_element *dmi;	/* intermediate */
	struct wifi_sta_events *sv;	/* sta events */
	int pending;
	object_t i1905;			/* i1905 UBUS object */
	object_t i1905_map;		/* i1905.map UBUS object */
	object_t topology;		/* topologyd UBUS object */
	uint8_t ieee1905_macaddr[6];	/* self node macaddr */
	struct cmdu_ackq cmdu_ack_q;
	//struct sta_eventlog assoc_stalog;
	void *subscriber;
	struct list_head framelist;	/* wifi_assoc_frame list */
	struct wifi_radio_events rv;	/* wifi.radio events */
	DELM_OBJECT_PLACEHOLDER
};

#define CMDU_TYPE_INVALID	-1

int decollector_add_device(struct wifi_network_device *remote_data, struct wifi_network_device *data);

extern int decollector_publish_object(struct decollector_private *p, const char *objname);
extern int decollector_register_events(struct decollector_private *p);
extern uint32_t decollector_lookup_object(struct decollector_private *p, const char *objname);


extern int decollector_dump_all(struct decollector_private *p, struct blob_buf *bb);
extern int decollector_dump2_all(struct decollector_private *p, struct blob_buf *bb);
extern int decollector_event_dump(struct decollector_private *p, struct blob_buf *bb);
extern int decollector_notify_events(struct decollector_private *p);
extern int decollector_collect_node(struct decollector_private *p, struct wifi_network_device *dev);
extern struct wifi_network_device *decollector_get_origin_dev(
		struct decollector_private *p, struct wifi_data_element *dm,
		uint8_t *origin);

extern int decollector_get_deagents(struct decollector_private *p, int *num, uint8_t **macaddrs);
extern int decollector_get_ieee1905id(struct decollector_private *p);

extern void decollector_getter_run(void *data);

extern void decollector_free_dm(struct wifi_data_element *dm);
extern void decollector_update_datamodel(struct decollector_private *p);

extern char *get_date(time_t *t, char *tbuf);
extern char *get_timestamp(time_t *t, char *tbuf);

// extern void *get_local_device_data(struct decollector_private *p);
// extern int decollector_add_neighbor_node(void *priv, void *remote_dev);
// extern int decollector_del_neighbor_node(void *priv, uint8_t *macaddr);
extern int self_node_topology_query(struct decollector_private *p);
extern void refresh_sta_timer_cb(atimer_t *t);
//extern int decollector_get_assoc_sta(struct decollector_private *p, const char *ifname,
//		uint8_t *sta_mac, int status_code, char *ev_tsp, uint8_t *bssid);
extern int decollector_get_unassoc_sta(struct decollector_private *p, uint8_t *sta_mac,
		int reason_code, char *ev_tsp);
extern int decollector_fill_sta_bssid(struct decollector_private *p, uint8_t *sta_mac, uint8_t *bssid);
extern int decollector_update_stalog_data(void *sta_log, const char *ifname, uint8_t *mac);
extern int prepare_autoconfig_search(struct decollector_private *priv, uint8_t band);
extern int decollector_subscribe_for_cmdus(struct decollector_private *priv);
extern uint8_t rssi_to_rcpi(int rssi);

#endif /* DECOLLECTOR_H */
